require 'luabit.bit'
require 'BinDecHex'

local function hex(str)
    str = Dec2Hex(str)
    
    if string.len(str) == 7 then
        str = '0' .. str
    end
    
    return str
end

local function ft(t, b, c, d)
    if t < 20 then
        return bit.bor(bit.band(b, c), bit.band(bit.bnot(b), d))
    elseif t < 40 then
        return bit.bxor(b, bit.bxor(c, d))
    elseif t < 60 then
        return bit.bor(bit.band(b, c), bit.bor(bit.band(b, d), bit.band(c, d)))
    else
        return bit.bxor(b, bit.bxor(c, d))
    end
end

local function kt(t)
    if t < 20 then
		return 1518500249;
	elseif t < 40 then
		return 1859775393;
	elseif t < 60 then
		return -1894007588;
	else
		return -899497514;
	end
end

local function safeAdd(x, y)
    local lsw = bit.band(x, 0xFFFF) + bit.band(y, 0xFFFF)
    local msw = bit.brshift(x, 16) + bit.brshift(y, 16) + bit.brshift(lsw, 16)
    return bit.bor(bit.blshift(msw, 16), bit.band(lsw, 0xFFFF))
end

local function zeroFill(a, b)
    local bin = Dec2Bin(a)
    
    if string.len(bin) < b then
        b = 0
    else
        bin = string.sub(bin, 1, string.len(bin) - b + 1)
    end
    
    local i = 0
    while i < b do
        bin = '0' .. bin
        i = i + 1
    end
    
    return tonumber(Bin2Dec(bin))
end

local function rol(num, cnt)
    return bit.bor(bit.blshift(num, cnt), zeroFill(num, 32 - cnt))
end

function sha1(str)
    local strlen = string.len(str)
    local n = bit.brshift(strlen + 8, 6) + 1
    local x = {}
    local i = 0
    
    while i < n * 16 do
        x[i] = 0
        i = i + 1
    end
    
    while i < strlen do
        local index = bit.brshift(i, 2)
        x[index] = bit.bor(x[index], bit.blshift(string.byte(string.sub(str, i + 1, 2)), 24 - (i % 4) * 8)) -- something wrong here
        i = i + 1
    end
    
    local index = bit.brshift(i, 2)
    x[index] = bit.bor(x[index], bit.blshift(0x80, 24 - (i % 4) * 8))
    x[n * 16 - 1] = strlen * 8
    
    local a =  1732584193;
	local b = -271733879;
	local c = -1732584194;
	local d =  271733878;
	local e = -1009589776;
	
	i = 0
	while i < #x do
	    local olda = a
	    local oldb = b
	    local oldc = c
	    local oldd = d
	    local olde = e
	    local j = 0
	    local w = {}
	    
	    while j < 80 do
	        if j < 16 then
	            w[j] = x[i + j]
            else
                w[j] = rol(bit.bxor(w[j - 2], bit.bxor(w[j - 8], bit.bxor(w[j - 14], w[j - 16]))), 1)
            end

            local t = safeAdd(safeAdd(rol(a, 5), ft(j, b, c, d)), safeAdd(safeAdd(e, w[j]), kt(j)))

            e = d
            d = c
            c = rol(b, 30)
            b = a
            a = t
	        j = j + 1
        end
	    
	    a = safeAdd(a, olda)
	    b = safeAdd(b, oldb)
	    c = safeAdd(c, oldc)
	    d = safeAdd(d, oldd)
	    e = safeAdd(e, olde)
	    i = i + 16
    end
    
    return hex(a) .. hex(b) .. hex(c) .. hex(d) .. hex(e)
end

--print(sha1('sdfjsiudhf334456'))